<?php

namespace Ivysaur;

/**
 * Tupac
 *
 * Static methods to pack two ints into one. Various packing modes are available
 *  with different properties. It may be possible to implement further packing
 *  modes based on any space-filling curve.
 */
class Tupac {
	
	const MODE_DIAGONAL = 0;
	const MODE_BINARY   = 1;
	
	const MODE__DEFAULT = self::MODE_DIAGONAL;
	
	protected function __construct() { }
	
	/**
	 * Pack two ints into one.
	 *
	 * @param int $a
	 * @param int $b
	 * @param int $mode Packing mode (see Tupac::MODE_* constants)
	 * @return int Packed result
	 */
	public static function pack($a, $b, $mode=self::MODE__DEFAULT) {
		switch ($mode) {
			case self::MODE_DIAGONAL: {
				return (
					( ($a + $b)*($a + $b + 1) )/2 + 1 + $a
				);
			} break;
			
			case self::MODE_BINARY: {
				$numbits = max( strlen(decbin($a)), strlen(decbin($b)) );
				$ret = 0;
				for($i = 0; $i < $numbits; ++$i) {
					$ret |= ($a & 1<<$i) << ($i);
					$ret |= ($b & 1<<$i) << ($i + 1);
				}
				return $ret;
			} break;
			
			default: {
				throw new \InvalidArgumentException(
					"Unrecognised mode parameter"
				);
			} break;
		}
	}
	
	/**
	 * Unpack an int into two.
	 *
	 * @param int $val Packed int
	 * @param int $mode Packing mode (see Tupac::MODE_* constants)
	 * @return int[] Array of two resulting numbers
	 */
	public static function unpack($val, $mode=self::MODE__DEFAULT) {
		switch ($mode) {
			case self::MODE_DIAGONAL: {
				// Find lowest y-intercept
				// FIXME: there should be a non-iterative solution to this
				$y = 0;
				$x = 0;
				$max = 0;
				for(;;) {
					$newmax = self::pack(0, $y, self::MODE_DIAGONAL);
					if ($newmax > $val) {
						break;
					}
					$max = $newmax;
					$y++;
				}
				$overshoot = $val - $max;				
				return [ $overshoot, $y - $overshoot - 1 ];
			} break;
			
			case self::MODE_BINARY: {
				$maxbits = strlen(decbin($val));
				$a = 0;
				$b = 0;
				for ($i = 0; $i < $maxbits; $i += 2) {
					$a |= ( ($val >>  $i   ) & 1 ) << $i/2;
					$b |= ( ($val >> ($i+1)) & 1 ) << $i/2;
				}
				return [$a, $b];
			} break;
			
			default: {
				throw new \InvalidArgumentException(
					"Unrecognised mode parameter"
				);
			} break;
		}
	}
	
}
